function [nvar,nxyzt,xyz,t]=v5dlocate(v5dstruc,var,xyza,tstr_or_tn);
%function [nvar,nxyzt,xyz,t]=v5dlocate(v5dstruc, var, xyza, ta)
%locate a variable in the vis5d structure. Returns nearest grid point to
% a specified location and time.
% v5dstruc  = name of the vis5d structure returned by vis5d('open','fileneame')
% var       = string contianing variable name (eg. 'Theta').
% xyza      = vector [xa, ya, za ] approximate x, y, z location.
% tstr_or_tn= time of desired v5d volume. This may be specified by a
%              time frame number (eg. 59) or as a matlab date 
%              string (eg.'01-Mar-1995 15:45:17') If only the time portion
%              of the string is specified (eg. '15:45:17' or '15:45') the date
%              portion defaults to the start date specified in the v5d file.
% nvar      = variable number in structure.
% nxyzt     = vector=[nx,ny,nz,nt]=indexes of point nearest to xyza,tstr_or_tn.
% xyz       = a vector [x,y,z] providing coordinates at point nx, ny, nz.
% t         = time at frame number nt. This is returned as a matlab
%              date string 
% If the requested coordinate is outside the Vis5D volume a warning is 
%printed and the nearest data point will be returned in xyz and nxyzt.

%Check date/time representation
if ischar(tstr_or_tn) == 1     %if time is specified as date string.
  [jnk,nchar]=size(tstr_or_tn);
  if nchar < 12  %if time is specified without date (eg. 12:30:41) 
     tstr_or_tn=[v5dstruc.times(1,1:12),tstr_or_tn];
  end;
end;

%check that the variable name was specified as a string;
if ischar(var) ~=1
   fprintf('****V5DLOCATE: FATAL ERROR-- var must be specified as a');
   fprintf(' string (eg. ''THETA'')****\n');
end;


xa=xyza(1);
ya=xyza(2);
za=xyza(3);
nvar=0;
nxyzt = [0 0 0 0];
xyz  = [nan nan nan];
t=nan;
%find location of requested variable 
for i=1:v5dstruc.nvars
  if strcmp(v5dstruc.vars(i).name, var)
      nvar=i;
      break;
  end; 
end;  
if nvar == 0 
   fprintf('\n****V5DLOCATE--Variable name not found in vis5d structure! \n\n')
   return;
end;


nxyzt(1) = 0;
for i=2:v5dstruc.xydims(1)
       %find x-gird points bracketing xa
       if (((v5dstruc.xaxis(i-1) >= xa) & (v5dstruc.xaxis(i) <= xa)) ...
               | ((v5dstruc.xaxis(i-1) <= xa) & (v5dstruc.xaxis(i) >= xa)))    
           %find which point is nearest.
           if abs(v5dstruc.xaxis(i-1) - xa) < abs(v5dstruc.xaxis(i)- xa)
               xyz(1) = v5dstruc.xaxis(i-1);  %x-value at nx
               nxyzt(1) = i-1;
               break;
           else
               xyz(1) = v5dstruc.xaxis(i);
               nxyzt(1) = i;
               break;
           end;
    end;
end;
if nxyzt(1) == 0
   if abs(v5dstruc.xaxis(1) - xa) < abs(v5dstruc.xaxis(v5dstruc.xydims(1))- xa)
       xyz(1) = v5dstruc.xaxis(1);  %x-value at nx
       nxyzt(1) = 1;
   else
       xyz(1) = v5dstruc.xaxis(v5dstruc.xydims(1));  %x-value at nx         
       nxyzt(1)  = v5dstruc.xydims(1);      
   end;
   fprintf('\n**********V5DLOCATE warning, x = %g', xa);
   fprintf(' not within vis5d volume**************\n');
   fprintf('**********substituting nearest x-value, new x = %g',xyz(1));
   fprintf(' *******************\n');
end;

for i=2:v5dstruc.xydims(2)
    if (((v5dstruc.yaxis(i-1) >= ya) & (v5dstruc.yaxis(i) <= ya)) ...
           | ((v5dstruc.yaxis(i-1) <= ya) & (v5dstruc.yaxis(i) >= ya)))
       if abs(v5dstruc.yaxis(i-1) - ya) < abs(v5dstruc.yaxis(i)- ya)
           xyz(2) = v5dstruc.yaxis(i-1);   %y-value at ny.
           nxyzt(2) = i-1;  
           break;
       else
           xyz(2) = v5dstruc.yaxis(i);
           nxyzt(2) = i;
           break;
       end;
    end;
end;
if nxyzt(2) == 0
   if abs(v5dstruc.yaxis(1) - ya) < abs(v5dstruc.yaxis(v5dstruc.xydims(2))- ya)
       xyz(2) = v5dstruc.yaxis(1);  
       nxyzt(2) = 1;
   else
       xyz(2) = v5dstruc.yaxis(v5dstruc.xydims(2));           
       nxyzt(2) = v5dstruc.xydims(2);      
   end;
   fprintf('\n**********V5DLOCATE warning, y = %g', ya);
   fprintf(' not within vis5d volume**************\n');
   fprintf('**********substituting nearest y-value, new y = %g',xyz(2));
   fprintf(' *******************\n');
end;

for i=2:v5dstruc.vars(nvar).zdim
%check if za is between this value of the zaxis grid point and the last value.
if(((v5dstruc.vars(nvar).zaxis(i-1)>=za)&(v5dstruc.vars(nvar).zaxis(i)<=za))...
    |((v5dstruc.vars(nvar).zaxis(i-1)<=za)&(v5dstruc.vars(nvar).zaxis(i)>=za)))
   
     if abs(v5dstruc.vars(nvar).zaxis(i-1) - za)...
                   < abs(v5dstruc.vars(nvar).zaxis(i)- za)
           xyz(3) = v5dstruc.vars(nvar).zaxis(i-1); %zvalue at nz. 
           nxyzt(3) = i-1;
           break;
     else
           xyz(3) = v5dstruc.vars(nvar).zaxis(i); %z value at nz.
           nxyzt(3) = i;
           break;
     end;
end;
end;

if nxyzt(3) == 0
   if (abs(v5dstruc.vars(nvar).zaxis(1) - za)...
                < abs(v5dstruc.vars(nvar).zaxis(v5dstruc.vars(nvar).zdim)- za))
       xyz(3) = v5dstruc.vars(nvar).zaxis(1);  
       nxyzt(3) = 1;
   else
       xyz(3) = v5dstruc.vars(nvar).zaxis(v5dstruc.vars(nvar).zdim);           
       nxyzt(3) = v5dstruc.vars(nvar).zdim;      
   end;

   fprintf('\n**********V5DLOCATE warning, z = %g', za);
   fprintf(' not within vis5d volume**************\n');
   fprintf('**********substituting nearest z-value, new z = %g',xyz(3));
   fprintf(' *******************\n');
end;

if v5dstruc.ntimes == 1                  % If only one time frame is structure.
  nxyzt(4)=1;
  t=v5dstruc.times(1,:);
elseif ischar(tstr_or_tn) == 0           % If time supplied as frame number.
   if ((tstr_or_tn > v5dstruc.ntimes) |(tstr_or_tn < 1))  
      fprintf('\n*******V5DLOCATE warning, time=%g',tstr_or_tn);
      fprintf(' is outside interval defined in Vis5D file*\n');
      fprintf('********substituting nearest time********************************************\n');
      if(tstr_or_tn <1) 
            nxyzt(4)=1;
            t=v5dstruc.times(1,:);
      else   
            nxyzt(4)=v5dstruc.ntimes;
            t= v5dstruc.times(v5dstruc.ntimes);
      end;
  else
      nxyzt(4)=tstr_or_tn;
      t=v5dstruc.times(tstr_or_tn,:);
  end;
else                                     % If time supplied as date string.
  time_a=datenum(tstr_or_tn);
  last_time=datenum(v5dstruc.times(1,:));
  for i=2:v5dstruc.ntimes
    time=datenum(v5dstruc.times(i,:));  
    if (((last_time <= time_a) & (time >= time_a))...
           | ((last_time >= time_a) & (time <= time_a)))
 
         if (abs(last_time - time_a) < abs(time- time_a))
              t = datestr(last_time);
              nxyzt(4) = i-1;  
              break;
         else
              t = datestr(time);
              nxyzt(4) = i;
              break;
         end;
    end;
    last_time=time;
  end;
end;

if nxyzt(4) == 0
      fprintf('\n*******V5DLOCATE warning, time= %s',tstr_or_tn);
      fprintf(' is outside interval defined');
      fprintf('********in Vis5D file, substituting nearest time*************************************\n');
   if time_a < datenum(v5dstruc.times(1,:))
       t=v5dstruc.times(1,:);
       nxyzt(4)=1;
   else
      t=v5dstruc.times(v5dstruc.ntimes,:);
      nxyzt(4)=v5dstruc.ntimes;
   end;
end;





