/*****************************************************************
 * vis5d-mex Version 0.13
 *
 * (c) Joseph Garcia, 2000, 2001 <jpgarcia@lidar.ssec.wisc.edu>
 * University of Wisconsin Lidar Group
 * Space Science and Engineering Center
 * Madison, WI, USA
 */

#include <stdio.h>
#ifdef V5DPLUS
#include <vis5d+/v5d.h>
#include <vis5d+/binio.h>
#else
#include "v5d.h"
#include "binio.h"
#endif
#include "mex.h"

#define VIS5DLAB_VERSION "0.13"
#define VIS5DLAB_RELEASE "2001.10.23"

#undef DEBUG
#undef VERBOSE

char *
getstring (const mxArray * a)
{
  int llen;
  char *line;
  
  if (!mxIsChar (a))
	return NULL;
  llen = mxGetN (a) * mxGetM (a) + 1;
  line = malloc (llen);
  mxGetString (a, line, llen);
  return line;
}

enum { ARG_COMMAND, ARG_NAME };
enum { ARG_VAR, ARG_X, ARG_Y, ARG_Z, ARG_T };
enum { FILE_INFO };
enum { RET_ARRAY };

#define VARSTRUCT_SIZE 5
static const char *varstruct_fields[VARSTRUCT_SIZE] = {
  "name", "units", "limits", "zdim", "zaxis"
};
enum { VAR_NAME, VAR_UNITS, VAR_EXTENTS, VAR_NLEV, VAR_LEVS };

#define INFOSTRUCT_SIZE 9
static const char *infostruct_fields[INFOSTRUCT_SIZE] = {
  "filename", "xydims", "ntimes", "xaxis",
  "yaxis", "times", "xyzunits", "nvars",
  "vars"
};
enum { V5DINFO_FILENAME, V5DINFO_DIMS, V5DINFO_TIMES, V5DINFO_XAXIS,
	   V5DINFO_YAXIS, V5DINFO_TIMEAXIS, V5DINFO_UNITS, V5DINFO_NVARS,
	   V5DINFO_VARS
};
static const char *axes_fields[3] = { "x", "y", "z" };

mxArray *
GetV5DVarsInfo (const v5dstruct * v5din)
{
  mxArray *mexout, *tmp;
  int c, d;
  
  mexout = mxCreateStructMatrix (1, v5din->NumVars, VARSTRUCT_SIZE, varstruct_fields);
  for (c = 0; c < v5din->NumVars; c++) {
	mxSetFieldByNumber (mexout, c, VAR_NAME, mxCreateString (v5din->VarName[c]));
	mxSetFieldByNumber (mexout, c, VAR_UNITS, mxCreateString (v5din->Units[c]));
	
	tmp = mxCreateDoubleMatrix (1, 2, mxREAL);
	mxGetPr (tmp)[0] = v5din->MinVal[c];
	mxGetPr (tmp)[1] = v5din->MaxVal[c];
	mxSetFieldByNumber (mexout, c, VAR_EXTENTS, tmp);
	
	tmp = mxCreateDoubleMatrix (1, 1, mxREAL);
	mxGetPr (tmp)[0] = v5din->Nl[c];
	mxSetFieldByNumber (mexout, c, VAR_NLEV, tmp);
	
	tmp = mxCreateDoubleMatrix (1, v5din->Nl[c], mxREAL);
	switch (v5din->VerticalSystem) {
	case 0:
	case 1:
	  for (d = 0; d < v5din->Nl[c]; d++)
		mxGetPr (tmp)[d] =
		  (double) (v5din->VertArgs[0] +
					v5din->VertArgs[1] * (d + v5din->LowLev[c]));
	  break;
	case 2:
	case 3:
	  for (d = 0; d < v5din->Nl[c]; d++)
		mxGetPr (tmp)[d] = (double) v5din->VertArgs[d + v5din->LowLev[c]];
	  break;
	default:
	  mexErrMsgTxt ("Unknown Vertical System");
	  return;
	}
	mxSetFieldByNumber (mexout, c, VAR_LEVS, tmp);
  }
  return mexout;
}

//in this order, they occur
#define MD_VERSION 0
#define MD_CLOSE   1
#define MD__OPEN   2
#define MD_INFO    4
#define MD_FILE    8
#define MD_OPEN    (MD_CLOSE|MD__OPEN|MD_INFO)

void
mexFunction (int nlhs, mxArray * plhs[], int nrhs, const mxArray * prhs[])
{
  static int file_open = 0;
  static char *filename = 0;
  static v5dstruct *v5d = 0;
  
  char **datestr;
  char *months[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
					   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
  };
  char *line = 0;
  
  int mode=0;
  void *data = 0;
  mxArray *tmp = 0;
  
  if (nrhs <= 0) {
	mexErrMsgTxt ("Invalid params");
	return;
  }
  
  if (line = getstring (prhs[ARG_COMMAND])) {
	if(!strcmp (line, "version"))
	  mode=MD_VERSION;
	else if(!strcmp (line, "info"))
	  mode=MD_INFO;
	else if(!strcmp (line, "open"))
	  mode=MD_OPEN;
	else if(!strcmp (line, "close"))
	  mode=MD_CLOSE;
	else if(!strcmp (line, "file"))
	  mode=MD_FILE;
	free(line);
	if (mode==MD_VERSION) {
	  if(nlhs!=0){
		mexErrMsgTxt ("too many return parameters");
		return;
	  }
	  if(nrhs!=1){
		mexErrMsgTxt ("too many parameters");
		return;
	  }
	  mexPrintf ("Vis5D-Mex %s\n", VIS5DLAB_VERSION);
	  mexPrintf ("Release %s\n", VIS5DLAB_RELEASE);
	  return;
	}
	if (mode&MD_CLOSE) {
	  if((mode==MD_CLOSE)&&(nrhs!=1)){
		mexErrMsgTxt ("too many parameters");
		return;
	  }
	  if (file_open) {
#ifdef VERBOSE
		mexPrintf ("Closing \'%s\'\n", filename);
#endif
		v5dCloseFile (v5d);
		free (filename);
		filename=0;
		plhs[RET_ARRAY] = mxCreateDoubleMatrix (1, 1, mxREAL);
		mxGetPr (plhs[RET_ARRAY])[0] = 0;
	  } else {
		plhs[RET_ARRAY] = mxCreateDoubleMatrix (1, 1, mxREAL);
		mxGetPr (plhs[RET_ARRAY])[0] = 1;
	  }
	  if(v5d){
		free (v5d);
		v5d=0;
	  }
	  file_open = 0;
	  if(mode==MD_CLOSE)
		return;
	}
	if (mode&MD__OPEN) {
	  if (nrhs < 2) {
		mexErrMsgTxt ("Open requires a filename");
		return;
	  }
	  if (!file_open) {
		if (filename)
		  free (filename);
		filename = getstring (prhs[ARG_NAME]);
		//mexPrintf ("opening \'%s\'\n", filename);
		v5d = v5dNewStruct ();
		if (!v5dOpenFile (filename, v5d)) {
		  mexErrMsgTxt ("Cannot open file");
		  free (v5d);
		  free (filename);
		  file_open = 0;
		  return;
		}
		mexPrintf ("Opened \'%s\'\n", filename);
		file_open = 1;
	  }
	}
	
	if (!file_open) {
	  mexErrMsgTxt ("No valid Vis5D file specified");
	  return;
	}
	
	if (mode&MD_INFO) {
	  int c;
	  double *dptr;
	  int julian_date, year, month, day;
	  
#ifdef DEBUG
	  mexPrintf ("Getting Info\n");
#endif
	  plhs[FILE_INFO] = mxCreateStructMatrix (1, 1, INFOSTRUCT_SIZE, infostruct_fields);
	  
	  mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_FILENAME, mxCreateString (filename));
	  
	  tmp = mxCreateDoubleMatrix (1, 2, mxREAL);
	  mxGetPr (tmp)[0] = v5d->Nc;
	  mxGetPr (tmp)[1] = v5d->Nr;
	  mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_DIMS, tmp);
	  
	  tmp = mxCreateDoubleMatrix (1, 1, mxREAL);
	  mxGetPr (tmp)[0] = v5d->NumTimes;
	  mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_TIMES, tmp);
	  
	  tmp = mxCreateDoubleMatrix (1, 1, mxREAL);
	  mxGetPr (tmp)[0] = v5d->NumVars;
	  mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_NVARS, tmp);
#ifdef DEBUG
	  mexPrintf ("Getting VAR Info\n");
#endif
	  
	  mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_VARS, GetV5DVarsInfo (v5d));
#ifdef DEBUG
	  mexPrintf ("done Getting var Info\n");
#endif
	  
	  tmp = mxCreateStructMatrix (1, 1, 3, axes_fields);
	  switch (v5d->Projection) {
	  case 0:
		mxSetFieldByNumber (tmp, 0, 0, mxCreateString (""));
		mxSetFieldByNumber (tmp, 0, 1, mxCreateString (""));
		break;
	  case 1:
	  case 4:
		mxSetFieldByNumber (tmp, 0, 0, mxCreateString ("deg"));
		mxSetFieldByNumber (tmp, 0, 1, mxCreateString ("deg"));
		break;
	  case 2:
	  case 3:
		mxSetFieldByNumber (tmp, 0, 0, mxCreateString ("km"));
		mxSetFieldByNumber (tmp, 0, 1, mxCreateString ("km"));
		break;
	  default:
		mexErrMsgTxt ("Unknown Projection System");
		return;
	  }
	  switch (v5d->VerticalSystem) {
	  case 0:
		mxSetFieldByNumber (tmp, 0, 2, mxCreateString (""));
		break;
	  case 1:
	  case 2:
		mxSetFieldByNumber (tmp, 0, 2, mxCreateString ("km"));
		break;
	  case 3:
		mxSetFieldByNumber (tmp, 0, 2, mxCreateString ("mb"));
		break;
	  default:
		mexErrMsgTxt ("Unknown Vertical System");
		return;
	  }
	  mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_UNITS, tmp);
	  
	  switch (v5d->Projection) {
	  case 0:
	  case 1:
		tmp = mxCreateDoubleMatrix (1, v5d->Nc, mxREAL);
		dptr = mxGetPr (tmp);
		for (c = 0; c < v5d->Nc; c++)
		  dptr[c] = -v5d->ProjArgs[1] + v5d->ProjArgs[3] * c;
		mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_XAXIS, tmp);
		tmp = mxCreateDoubleMatrix (1, v5d->Nr, mxREAL);
		dptr = mxGetPr (tmp);
		for (c = 0; c < v5d->Nr; c++)
		  dptr[c] = v5d->ProjArgs[0] - v5d->ProjArgs[2] * c;
		mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_YAXIS, tmp);
		break;
	  case 2:
		tmp = mxCreateDoubleMatrix (1, v5d->Nc, mxREAL);
		dptr = mxGetPr (tmp);
		for (c = 0; c < v5d->Nc; c++)
		  dptr[c] = v5d->ProjArgs[5] * (c - (v5d->Nc + 1) / 2);
		mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_XAXIS, tmp);
		tmp = mxCreateDoubleMatrix (1, v5d->Nr, mxREAL);
		dptr = mxGetPr (tmp);
		for (c = 0; c < v5d->Nr; c++)
		  dptr[c] = v5d->ProjArgs[5] * (c - (v5d->Nc / 2));
		mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_YAXIS, tmp);
		break;
	  case 3:
		tmp = mxCreateDoubleMatrix (1, v5d->Nc, mxREAL);
		dptr = mxGetPr (tmp);
		for (c = 0; c < v5d->Nc; c++)
		  dptr[c] = v5d->ProjArgs[4] * (c - v5d->ProjArgs[3]);
		mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_XAXIS, tmp);
		tmp = mxCreateDoubleMatrix (1, v5d->Nr, mxREAL);
		dptr = mxGetPr (tmp);
		for (c = 0; c < v5d->Nr; c++)
		  dptr[c] = v5d->ProjArgs[4] * (c - v5d->ProjArgs[2]);
		mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_YAXIS, tmp);
		break;
	  case 4:
		tmp = mxCreateDoubleMatrix (1, v5d->Nc, mxREAL);
		dptr = mxGetPr (tmp);
		for (c = 0; c < v5d->Nc; c++)
		  dptr[c] = v5d->ProjArgs[3] * c * 100;
		mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_XAXIS, tmp);
		tmp = mxCreateDoubleMatrix (1, v5d->Nr, mxREAL);
		dptr = mxGetPr (tmp);
		for (c = 0; c < v5d->Nr; c++)
		  dptr[c] = v5d->ProjArgs[2] * (v5d->Nr - c - 1) * 100;
		mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_YAXIS, tmp);
		break;
	  default:
		mexErrMsgTxt ("Unknown Projection System");
		return;
	  }
#ifdef DEBUG
	  mexPrintf ("done axis Info\n");
#endif
	  
	  datestr = (char **) malloc (sizeof (char *) * v5d->NumTimes);
	  for (c = 0; c < v5d->NumTimes; c++) {
		month = 1;
		julian_date = v5d->DateStamp[c];
		day = julian_date % 1000;
		year = julian_date / 1000;
		if (year > 70)	/* Y2K WARNING */
		  year += 1900;
		else
		  year += 2000;
		while (julian_date != 0)
		  switch (month) {
		  case (1):
		  case (3):
		  case (5):
		  case (7):
		  case (8):
		  case (10):
		  case (12):
			if (day > 31) {
			  month++;
			  day -= 31;
			} else
			  julian_date = 0;
			break;
		  case (4):
		  case (6):
		  case (9):
		  case (11):
			if (day > 30) {
			  month++;
			  day -= 30;
			} else
			  julian_date = 0;
			break;
		  case (2):
			{
			  int daycount = (28 + (((((year % 4) == 0)
									  && ((year % 100) != 0)) ||	/* is div by 4, & !100 */
									 ((year % 400) == 0)) ? 1 : 0));
			  
			  if (day > daycount) {
				month++;
				day -= daycount;
			  } else
				julian_date = 0;
			  break;
			}
		  }
		datestr[c] = (char *) malloc (sizeof (char) * 25);
		if(v5d->TimeStamp[c] <= 235959)
		  sprintf (datestr[c], "%02i-%s-%04i %02i:%02i:%02i", day,
				   months[month - 1], year, v5d->TimeStamp[c] / 10000,
				   (v5d->TimeStamp[c] / 100) % 100, v5d->TimeStamp[c] % 100);
		else
		  sprintf (datestr[c], "%02i-%s-%04i %02i:%02i:%02i.%03i", day,
				   months[month - 1], year, v5d->TimeStamp[c] / 10000000,
				   (v5d->TimeStamp[c] / 100000) % 100, (v5d->TimeStamp[c]/1000) % 100, v5d->TimeStamp[c] % 1000);
	  }
	  mxSetFieldByNumber (plhs[FILE_INFO], 0, V5DINFO_TIMEAXIS,
						  mxCreateCharMatrixFromStrings (v5d->NumTimes, (const char **) datestr));
	  for (c = 0; c < v5d->NumTimes; c++)
		free (datestr[c]);
	  free (datestr);
#ifdef DEBUG
	  mexPrintf ("done time axis\n");
#endif
	}
	if (mode==MD_FILE) {
	  if(nrhs!=1){
		mexErrMsgTxt ("too many parameters");
		return;
	  }
	  if (file_open) 
		plhs[RET_ARRAY] = mxCreateString (filename);
	  else
		plhs[RET_ARRAY] = mxCreateString ("");
	  return;
	}
  } else {
	int req[5], dims[5], loc[5], maxDims[5];	/* X,Y,Z,T,Var */
	int i;
	void *comp_grid;
	float *grid;
	double *mexptr;
	mxArray *mexout;
	float ga[MAXLEVELS], gb[MAXLEVELS];
	
	if (!file_open) {
	  mexErrMsgTxt ("No valid Vis5D file specified");
	  return;
	}
	if (nrhs != 5) {
	  mexErrMsgTxt ("Invalid params");
	  return;
	}
	if(nlhs!=1){
	  mexErrMsgTxt ("needs a return parameter");
	  return;
	}
	req[4] = (int) mxGetScalar (prhs[ARG_VAR]);
	if ((req[4] <= 0) || (req[4] > v5d->NumVars)) {
	  mexErrMsgTxt ("Invalid Variable");
	  return;
	}
	req[4]--;	/* Matlab = [ 1 , NumVars ], Vis5D = [0 , NumVars-1] */
	maxDims[0] = v5d->Nc;
	maxDims[1] = v5d->Nr;
	maxDims[2] = v5d->Nl[req[4]];
	maxDims[3] = v5d->NumTimes;
	maxDims[4] = v5d->NumVars;
	
	for (i = 0; i < 4; i++)
	  if (mxGetN (prhs[i + ARG_X]) * mxGetN (prhs[i + ARG_X]) != 1) {
		req[i] = 0;	/* offset of 0 */
		dims[i] = maxDims[i];
	  } else {
		req[i] = ((int) mxGetScalar (prhs[i + ARG_X])) - 1;
		dims[i] = 1;
		if ((req[i] < 0) || (req[i] >= maxDims[i])) {
		  mexPrintf
			("Value \'%i\' for %s is out of [ 1, %i] bounds\n",
			 req[i] + 1,
			 (i <
			  2 ? (i == 0 ? "X-Axis" : "Y-Axis") : (i == 2 ? "Z-Axis" : (i == 3 ? "Time-Axis" : "Variable"))), maxDims[i]);
		  mexErrMsgTxt ("Aborting");
		  return;
		}
	  }
	/* Matlab Array */
	mexout = mxCreateNumericArray (4, dims, mxDOUBLE_CLASS, mxREAL);
	mexptr = mxGetPr (mexout);
	/*allocate */
	comp_grid = (void *) malloc (maxDims[0] * maxDims[1] * maxDims[2] * v5d->CompressMode);
	grid = (float *) malloc (maxDims[0] * maxDims[1] * maxDims[2] * sizeof (float));
	
#ifdef DEBUG
	mexPrintf ("Begin\n");
#endif
	for (loc[3] = 0; loc[3] < dims[3]; loc[3]++) {
	  /* open */
	  if (!v5dReadCompressedGrid (v5d, loc[3] + req[3], req[4], ga, gb, comp_grid)) {
		mexErrMsgTxt ("Error reading Vis5D compressed grid");
		return;
	  }
	  v5dDecompressGrid (v5d->Nr, v5d->Nc, v5d->Nl[req[4]], v5d->CompressMode, comp_grid,
						 ga, gb, grid);
	  for (loc[2] = 0; loc[2] < dims[2]; loc[2]++)
		for (loc[0] = 0; loc[0] < dims[0]; loc[0]++)
		  for (loc[1] = 0; loc[1] < dims[1]; loc[1]++)
			/* assign value */
			if (IS_MISSING (grid[(loc[1] + req[1]) +
								(loc[0] + req[0]) * maxDims[1] + (loc[2] +
																  req[2]) * maxDims[1] * maxDims[0]]))
			  mexptr[loc[0] + loc[1] * dims[0] +
					loc[2] * dims[1] * dims[0] +
					loc[3] * dims[1] * dims[0] * dims[2]] = mxGetNaN ();
			else
			  mexptr[loc[0] + loc[1] * dims[0] +
					loc[2] * dims[1] * dims[0] +
					loc[3] * dims[1] * dims[0] * dims[2]] =
				(double) grid[(loc[1] + req[1]) +
							 (loc[0] + req[0]) * maxDims[1] + (loc[2] +
															   req[2]) * maxDims[1] * maxDims[0]];
	}
	/*dealloc */
	free (comp_grid);
	free (grid);
	mexCallMATLAB(1,plhs, 1, &mexout, "squeeze");
#ifdef DEBUG
	mexPrintf ("Returned a [ %i x %i x %i x %i ] matrix\n", dims[0], dims[1], dims[2],
			   dims[3]);
#endif
  }
#ifdef DEBUG
  mexPrintf ("--DONE--\n");
#endif
  return;
}
